#include <stdio.h>
#include <string.h>
#include "../../brl.mod/blitz.mod/blitz.h"


/*
	Object CopyObject(Object instance);
	
	Copy an Object of any type.
	This will copy references as they
	are. So references in the copied
	instance will point at the same
	structures as references from the
	source instance.
*/

BBObject* CopyObject(BBObject *inst){
	int size;
	BBObject *copy;
	if(inst == &bbNullObject) return inst;
	size = inst->clas->instance_size;
	copy = bbObjectNew(inst->clas);
	memcpy(copy, inst, size);
	return copy;
}

/*		
	Object CopyObjectTree(Object instance);
	
	Copy whole instance recursively and try
	to avoid following references more than
	once.
	It parses the "debug_scope" which is a
	kind of array that saves all names of
	fields, methods etc., their types and
	offset.

	If BlitzMax is ever to support 64 bit
	this code will probably need to be
	modified...
*/

#define REFSTART	128
#define REFINC		128

struct RefList{
	void **ptr;
	unsigned int size, count;
};

void addref(struct RefList *refs, void *ptr, void *ptr2){
	refs->ptr[refs->count++] = ptr;
	refs->ptr[refs->count++] = ptr2;
	if(refs->count == refs->size){
		refs->size += REFINC;
		refs->ptr = realloc(refs->ptr, refs->size);
	}
}

void* findref(struct RefList *refs, void *ptr){
	int i;
	for(i = 0; i < refs->count; i += 2){
		if(refs->ptr[i] == ptr) return refs->ptr[i+1];
	}
	return NULL;
}

BBObject* __CopyObjectTree(BBObject *, struct RefList *);

BBObject* CopyObjectTree(BBObject *inst){
	struct RefList *refs = malloc(sizeof(struct RefList));
	BBObject *copy;
	
	refs->ptr = malloc(REFSTART);
	refs->size = REFSTART;
	refs->count = 0;
	
	addref(refs, &bbNullObject, &bbNullObject);
	
	copy = __CopyObjectTree(inst, refs);
	free(refs);
	
	return copy;
}

BBObject* __CopyObjectTree(BBObject *inst, struct RefList *refs){
	BBObject *copy = CopyObject(inst);
	int offset;
	char *name, *type;
	BBDebugScope *dscope;
	BBDebugDecl *info;
	void *ref;
	BBString *string;
	
	if(copy == &bbNullObject) return copy;
	
	addref(refs, inst, copy);

	dscope = (BBDebugScope*) inst->clas->debug_scope;
	info = dscope->decls;
	
	for(;;){
		if(info->kind == BBDEBUGDECL_END) break;

		if(info->kind == BBDEBUGDECL_FIELD){
		
			name = info->name;
			type = info->type_tag;
			offset = info->field_offset;
								
			if(type[0] == ':'){
				ref = findref(refs, *((void**) (((void*) inst) + offset)));
				if(ref){
					*((void**) (((void*) copy) + offset)) = ref;
				}else{
					*((void**) (((void*) copy) + offset)) = __CopyObjectTree(*((void**) (((void*) inst) + offset)), refs);
					addref(refs, *((void**) (((void*) inst) + offset)), *((void**) (((void*) copy) + offset)));
				}
			}else if(type[0] == '$'){
				string = *(BBString**)(((int) inst) + offset);
				ref = findref(refs, string);
				if(ref){
					*((void**) (((void*) copy) + offset)) = ref;
				}else{
					*((BBString**) (((void*) copy) + offset)) = bbStringSlice(string, 0, string->length);
					addref(refs, string, *((BBString**) (((void*) copy) + offset)));
				}
			}
			
			
		}
		
		info += 1;
	}
	return copy;
}

void* ObjectPtr(void *x){
	return x;
}

char* ParseClass(void *oex, BBClass *clas, void (*cb)(void*, int, char*, char*, int)){
	BBDebugScope *dscope;
	BBDebugDecl *info;

	dscope = (BBDebugScope*) clas->debug_scope;
	info = dscope->decls;
	
	for(;;){
		if(info->kind == BBDEBUGDECL_END) break;

		switch(info->kind){
		case BBDEBUGDECL_CONST:
		case BBDEBUGDECL_GLOBAL:
		case BBDEBUGDECL_FIELD:
		case BBDEBUGDECL_TYPEMETHOD:
		case BBDEBUGDECL_TYPEFUNCTION:
								
			cb(oex, info->kind, info->name, info->type_tag, info->field_offset);
			
			break;
		}
		
		info += 1;
	}
	
	return dscope->name;	
}

char* ParseObject(void *oex, BBObject *inst, void (*cb)(void*, int, char*, char*, int)){
	return ParseClass(oex, inst->clas, cb);
}
